/**********************************************************************
 *** fonction de test qui appelle la fonction "predit_octet" avec des *
 *** paramètres pris sur la ligne de commande                         *
 *** pour le TP3 d'info-614 : RC4, WEP et attaque FMS                 *
 *** Pierre Hyvernat                                                  *
 **********************************************************************/

#include "tp3.h"
#include <getopt.h>
#include <sys/time.h>

/* la clé WEP du fichier rca4.c */
extern octet *cle_wep;

/* affiche une clé WEP, en hexadécimal */
void affiche_cle(octet *key)
{
  for (int i = 0; i < taille_cle - taille_IV; i++)
    printf("%02x  ", key[i]);
}

/* affiche l'aide */
void affiche_aide(char *exe)
{
  printf("Usage : %s [options]\n", exe);
  printf("\n");
  printf("  --nb-tests=n, -t n\n");
  printf("      nombre de tests à faire (défaut : 1)\n");
  printf("      (incompatible avec \"-k ...\")\n");
  printf("  --taille-cle=n, -l n\n");
  printf("      taille en octet des clés (défaut : 16)\n");
  printf("  --taille-IV=n, -i n\n");
  printf("      taille en octet des vecteurs d'initialisation (défaut : 3)\n");
  printf("  --nb-IV, -V n\n");
  printf("      nombre de vecteurs d'initialisations à considérer pour chaque\n");
  printf("      octet\n");
  printf("  --nb-IV-faibles=n, -F n\n");
  printf("      nombre de vecteurs d'initialisations faibles à considérer pour\n");
  printf("      chaque octet\n");
  printf("  --cle=\"...\", -k \"...\"\n");
  printf("      essaie de cracker la clé donnée en argument sous la forme\n");
  printf("      \".. .. .. ..\" (en hexadécimal)\n");
  printf("      (incompatible avec \"-t ...\"\n");
  printf("  -v\n");
  printf("      augmente la verbosité des tests\n");
  printf("  -h\n");
  printf("      affiche ce message\n");

}

/* génère une cle à partir d'une chaîne de caractères représentant une clé en
 * hexadécimal... */
int string_to_key(char *cle)
{
  unsigned o;
  int i, j;
  for (i = 0; i < taille_cle - taille_IV && 3 * i < (int)strlen(cle); i++) {
    sscanf(&cle[3 * i], "%02x ", &o);
    cle_wep[i] = (octet) o;
  }
  for (j = i; j < taille_cle - taille_IV; j++)
    cle_wep[j] = 0;
  return i;
}

int main(int argc, char *argv[])
{
  /* initialisation générateur aléatoire */
  struct timeval temps;
  gettimeofday(&temps, NULL);
  srand(temps.tv_usec);

  int nb_tests = 1;                    // nombre de tests à faire (clés à cracker)
  int nb_IV_total = 0;                 // nombre de vecteurs à considérer pour chaque octet de la clé
  int nb_IV_faibles = 0;               // nombre de vecteurs faibles à considérer pour chaque octet de la clé
  int cle_fournie = 0;                 // pour savoir si la clé est fournie en argument
  extern int verbose;                  // variable globale du fichier tp3-nom-prenom.c
  verbose = 0;
  taille_cle = 16;
  taille_IV = 3;

  static const struct option long_options[] = {
    {"nb-tests", required_argument, NULL, 't'},
    {"taille-cle", required_argument, NULL, 'l'},
    {"taille-IV", required_argument, NULL, 'i'},
    {"nb-IV", required_argument, NULL, 'V'},
    {"nb-IV-faibles", required_argument, NULL, 'F'},
    {"cle", required_argument, NULL, 'k'},
    {NULL, 0, NULL, 0}
  };

  int opt;
  while ((opt = getopt_long(argc, argv, "ht:l:i:V:F:vk:", long_options, NULL)) != -1) {
    switch (opt) {
    case 'h':
      affiche_aide(argv[0]);
      return (0);
      break;
    case 't':
      nb_tests = atoi(optarg);
      break;
    case 'l':
      taille_cle = atoi(optarg);
      break;
    case 'i':
      taille_IV = atoi(optarg);
      break;
    case 'V':
      nb_IV_total = atoi(optarg);
      break;
    case 'F':
      nb_IV_faibles = atoi(optarg);
      break;
    case 'v':
      verbose++;
      break;
    case 'k':
      nouvelle_cle_wep_aleatoire();    // juste pour allouer la mémoire...
      int c = string_to_key(optarg);
      if (c < taille_cle - taille_IV && verbose > 0)
        printf("Attention : la clé fournie est trop courte, elle a été complétée par des \"00\".\n");
      if (c > taille_cle - taille_IV && verbose > 0)
        printf("Attention : la clé fournie est trop longue, elle a été tronquée.\n");
      cle_fournie = 1;
      break;
    default:                          /* '?' */
      printf("option non reconnue...\n\n");
      affiche_aide(argv[0]);
      return (-1);
    }
  }

  octet cle_partielle[taille_cle - taille_IV];      // la clé que l'on est en train de reconstruire
  octet diff[taille_cle - taille_IV];               // le tableau des différences entre la clé reconstruite et la clé réelle

  int i, k;
  int nb_erreurs = 0;
  int OK;

  if (cle_fournie == 1) {
    /* on réinitialise la clé partielle (normalement non nécessaire) */
    for (k = 0; k < taille_cle - taille_IV; k++)
      cle_partielle[k] = 0;

    /* on essaie de cracker tous les octets de la clé, un par un */
    for (k = 0; k < taille_cle - taille_IV; k++) {
      predit_octet(k, cle_partielle, nb_IV_faibles, nb_IV_total);
    }

    /* on compare la clé trouvée avec la vraie clé */
    int OK = 1;
    for (k = 0; k < taille_cle - taille_IV; k++) {
      diff[k] = cle_partielle[k] ^ cle_wep[k];
      if (diff[k] != 0) {
        OK = -k;
        nb_erreurs++;
        break;
      }
    }

    if (OK > 0) {
      printf("OK, clé =  ");
      affiche_cle(cle_wep);
      printf("\n");
    } else {
      printf("OOPS, erreur à l'octet %d.\n", -OK);
      printf("  - clé WEP     :  ");
      affiche_cle(cle_wep);
      printf("\n");
      printf("  - clé trouvée :  ");
      affiche_cle(cle_partielle);
      printf("\n");
    }

    return 0;
  }

  for (i = 1; i <= nb_tests; i++) {

    /* on tire une clé */
    nouvelle_cle_wep_aleatoire();

    /* on réinitialise la clé partielle (normalement non nécessaire) */
    for (k = 0; k < taille_cle - taille_IV; k++)
      cle_partielle[k] = 0;

    /* on essaie de cracker tous les octets de la clé, un par un */
    for (k = 0; k < taille_cle - taille_IV; k++)
      predit_octet(k, cle_partielle, nb_IV_faibles, nb_IV_total);

    /* on compare la clé trouvée avec la vraie clé */
    OK = 1;
    for (k = 0; k < taille_cle - taille_IV; k++) {
      diff[k] = cle_partielle[k] ^ cle_wep[k];
      if (diff[k] != 0) {
        OK = -k;
        nb_erreurs++;
        break;
      }
    }

    if (verbose == 0 && nb_tests > 2) {
      if (OK > 0)
        printf(".");
      else
        printf("+");
      fflush(NULL);
    }

    if (verbose > 0) {
      if (OK > 0) {
        printf("* test %d : OK\n", i);
        if (verbose > 1) {
          printf("  clé :  ");
          affiche_cle(cle_wep);
          printf("\n");
        }
      } else {
        printf("* test %d : OOPS, erreur à l'octet %d de la clé.\n", i,
               -OK);
        if (verbose > 1) {
          printf("  clé WEP     :  ");
          affiche_cle(cle_wep);
          printf("\n");
          printf("  clé trouvée :  ");
          affiche_cle(cle_partielle);
          printf("\n");
        }
      }
      printf("\n");
    }
  }

  printf("\nRésumé :\n");
  printf("    - clés de %d octets (IV) + %d octets\n", taille_IV, taille_cle - taille_IV);
  if (nb_IV_total > 0 && nb_IV_faibles < 1)
    printf("    - collecte de %d vecteurs par octet\n", nb_IV_total);
  if (nb_IV_total < 1 && nb_IV_faibles > 0)
    printf("    - collecte de %d vecteurs faibles par octet\n", nb_IV_faibles);
  if (nb_IV_total < 1 && nb_IV_faibles < 1)
    printf("    - utilisation des 256 vecteurs de la forme (n+3, 255, ...)\n");
  if (nb_IV_total > 0 && nb_IV_faibles > 0)
    printf("    - collecte de %d vecteurs faibles parmi au plus %d vecteurs par octet\n", nb_IV_faibles, nb_IV_total);
  printf("    - %d clés sur %d ont été craquées avec succès (%.4f%%)\n", nb_tests - nb_erreurs, nb_tests, (nb_tests - nb_erreurs) * 100.0 / nb_tests);
  printf("\n");

  return 0;
}
